#!/usr/bin/perl -w

# Sync PPA branches from the development branches
# (c) 2008-2010, Fabien Tassin <fta@sofaraway.org>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

use strict;
use warnings;

use Getopt::Std;
use Data::Dumper;
use File::Spec;

# Defaults
my $conf = {
  'dist'          => `lsb_release -c -s`,
  'alt_dists'     => [ 'intrepid', 'hardy' ], # other than $dist, will be taggued as ~dist
  'who'           => `bzr whoami`,
  'watchers'      => `bzr whoami`,
  'dput_target'   => "ppa", # from ~/.dput.cf
  'iteration_tag' => "ppa", # "xxx" => -0ubuntu1~xxx1
  'ppa_branchtag' => "ppa", # "xxx" => package.head.xxx
  'gpg_key'       => undef,
  'test_suffix'   => "test",
  'build_dir'     => "$ENV{'HOME'}/ppa",
  'tarballs_dir'  => "../tarballs", # absolute dir or relative to the branch dir
};

# List of conf files to consider, the 1st found wins
my $ppa_confs = [ "$ENV{'HOME'}/.ppabot.conf", "/etc/ppabot.conf" ];
my $ppa_pkgs  = [ "$ENV{'HOME'}/.ppabot-pkgs.conf", "/etc/ppabot-pkgs.conf" ];

my $opt = {};
getopts('d:hitrAfc:p:P', $opt) || &Usage();
&Usage() if $$opt{'h'};

unshift @$ppa_confs, $$opt{'c'} if defined $$opt{'c'};
unshift @$ppa_pkgs,  $$opt{'p'} if defined $$opt{'p'};

for my $ppa_conf (@$ppa_confs) {
  read_conf(\$conf, $ppa_conf, 1), last if -e $ppa_conf;
}

my $tarballs_dir = File::Spec->rel2abs("../" . $conf->{'tarballs_dir'});
if ($conf->{'tarballs_dir'} =~ m!^/!) {
	my $tarballs_dir = File::Spec->rel2abs($conf->{'tarballs_dir'});
}
my $build_dir = File::Spec->rel2abs("../" . $conf->{'build_dir'});
if ($conf->{'build_dir'} =~ m!^/!) {
	$build_dir = $conf->{'build_dir'};
}

my $key = defined $conf->{'gpg_key'} ? "-k$conf->{'gpg_key'}" : "";
my $bzr_cmds = {
  "ppa"  => [ "bd", "--merge",   "--build-dir=$build_dir", "--orig-dir=$tarballs_dir", "--builder=dpkg-buildpackage -rfakeroot -S -si $key" ],
  "ppa2" => [ "bd", "--merge",   "--build-dir=$build_dir", "--orig-dir=$tarballs_dir", "--builder=dpkg-buildpackage -rfakeroot -S -sd $key" ],
  "ppan" => [ "bd", "--native",  "--build-dir=$build_dir", "--orig-dir=$tarballs_dir", "--builder=dpkg-buildpackage -rfakeroot -S -sa $key" ],
  "bdn"  => [ "bd", "--native",  "--build-dir=$build_dir.tmp", "--orig-dir=$tarballs_dir" ],
  "bdm"  => [ "bd", "--merge",   "--build-dir=$build_dir.tmp", "--orig-dir=$tarballs_dir" ],
};

my $pkgs;
for my $ppa_pkg (@$ppa_pkgs) {
  read_conf(\$pkgs, $ppa_pkg, 0), last if -e $ppa_pkg;
}
# Initialize the 'ppa' branch names, if missing
for my $pkg (keys %{$pkgs->{'pkgs'}}) {
  $pkgs->{'pkgs'}{$pkg}{'ppa'} = $pkgs->{'pkgs'}{$pkg}{'branch'} . "." . $conf->{'ppa_branchtag'}
    unless defined $pkgs->{'pkgs'}{$pkg}{'ppa'};
}

die "You must create a ppabot-pkgs.conf file. Abort.\n" unless defined $pkgs;

############################################################################

sub Usage {
  die "Usage: $0 [-d archive] [-c conf] [-p pkgconf] [-f] [-i] [-P] [-t] [-h] [-A] [pkg1] [pkg2] [...]\n" .
    "  -h           this help screen\n" .
    "  -c conf      use 'conf' as alternative conf file\n" .
    "  -p pkgconf   use 'pkgconf' as alternative package conf file\n" .
    "  -d archive   push to an alternative archive (default: $conf->{'dput_target'})\n" .
    "  -t           use test branches, if necessary branched off the configured ones\n" .
    "  -A           don't do alternative dists\n" .
    "  -i           initialize the PPA branches and quit\n" .
    "  -f           force the merge, even if it's not needed\n" .
    "  -P           do the dPut at the end\n" .
    "  -r           revert the last sync\n";
}

chomp($conf->{'who'});
chomp($conf->{'dist'});
my $force_test = 0;
my $revert = 0;
undef $conf->{'alt_dists'} if $$opt{'A'};
$force_test = 1 if $$opt{'t'};
$revert = 1 if $$opt{'r'};
my $force = defined $$opt{'f'} ? 1 : 0;
my $init = defined $$opt{'i'} ? 1 : 0;
my $dput_dist = defined $$opt{'d'} ? $$opt{'d'} : $conf->{'dput_target'};
$conf->{'do_dput'} = defined $$opt{'P'} ? 1 : 0;

if ($#ARGV >= 0) {
  splice @{$pkgs->{'order'}}, 0, $#{$pkgs->{'order'}} + 1, @ARGV;
}
else {
  # Drop all no_auto packages from the 'order' list
  my $res = [];
  for my $pkg (@{$pkgs->{'order'}}) {
    push @$res, $pkg unless defined $pkgs->{'pkgs'}{$pkg}{'no_auto'} &&
      ($pkgs->{'pkgs'}{$pkg}{'no_auto'} eq '1' ||
       $pkgs->{'pkgs'}{$pkg}{'no_auto'} eq 'true');
  }
  $pkgs->{'order'} = $res;
}

my $rpkgs = {};
map { $$rpkgs{$_} = 1 } @{$pkgs->{'order'}};

# Create main Daily branches if they are missing
for my $pkg (@{$pkgs->{'order'}}) {
  die "'$pkg' not found in the package configuration files. Abort\n"
    unless defined $pkgs->{'pkgs'}{$pkg};
  if (! -d "$pkgs->{'pkgs'}{$pkg}{'branch'}/.bzr" &&
    defined $pkgs->{'pkgs'}{$pkg}{'lp'}) {
    print "*INFO* no $pkgs->{'pkgs'}{$pkg}{'branch'} branch found. " .
          "Fetching it from $pkgs->{'pkgs'}{$pkg}{'lp'} into $pkgs->{'pkgs'}{$pkg}{'branch'}..\n";
    `bzr branch $pkgs->{'pkgs'}{$pkg}{'lp'} $pkgs->{'pkgs'}{$pkg}{'branch'}`;
  }
  die "No branch $pkgs->{'pkgs'}{$pkg}{'branch'} found in the current directory. " .
    "You may have to branch/clone it there. Stop.\n"
      unless -d "$pkgs->{'pkgs'}{$pkg}{'branch'}/.bzr";
  die "Project $pkg is not configured. stop.\n" unless defined $pkgs->{'pkgs'}{$pkg};
  my $ppa = $pkgs->{'pkgs'}{$pkg}{'ppa'};
  if (! -d $ppa) {
    print "*INFO* no $ppa branch found. Creating it..\n";
    `bzr branch $pkgs->{'pkgs'}{$pkg}{'branch'} $ppa`;
  }
}

# Create PPA branches for all distributions if they are missing
for my $d ($conf->{'dist'}, @{$conf->{'alt_dists'}}) {
  for my $pkg (@{$pkgs->{'order'}}) {
    die "Project $pkg is not configured. stop.\n" unless defined $pkgs->{'pkgs'}{$pkg};
    my $ppa = $pkgs->{'pkgs'}{$pkg}{'ppa'};
    if (! -d "$ppa.$d") {
      print "*INFO* no $ppa.$d branch found. Creating it..\n";
      `bzr branch $ppa "$ppa.$d"`;
    }
  }
}

if ($force_test) {
  for my $pkg (@{$pkgs->{'order'}}) {
    die "Project $pkg is not configured. stop.\n" unless defined $pkgs->{'pkgs'}{$pkg};
    my $branch = $pkgs->{'pkgs'}{$pkg}{'branch'};
    my $ppa = $pkgs->{'pkgs'}{$pkg}{'ppa'};
    if (! -d "$branch.$conf->{'test_suffix'}") {
      print "*INFO* no $branch.$conf->{'test_suffix'} branch found. Creating it..\n";
      `bzr branch $branch "$branch.$conf->{'test_suffix'}"`;
      $pkgs->{'pkgs'}{$pkg}{'branch'} = "$branch.$conf->{'test_suffix'}";
    }
    if (! -d "$ppa.$conf->{'test_suffix'}") {
      print "*INFO* no $ppa.$conf->{'test_suffix'} branch found. Creating it..\n";
      `bzr branch $ppa "$ppa.$conf->{'test_suffix'}"`;
      $pkgs->{'pkgs'}{$pkg}{'ppa'} = "$ppa.$conf->{'test_suffix'}";
    }
  }
}

exit 0 if $init;
my $versions = {};
my $cids = {};
my $date = `date -R`;
chomp($date);
my $dput = [];

for my $pkg (keys %{$pkgs->{'pkgs'}}) {
  next unless defined $pkgs->{'pkgs'}{$pkg};
  if (defined $$rpkgs{$pkg} && !$revert) {
    check_committed($pkg, $pkgs->{'pkgs'}{$pkg}{'branch'});
    check_committed($pkg, $pkgs->{'pkgs'}{$pkg}{'ppa'});
  }
  $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}} = get_pkg_version($pkgs->{'pkgs'}{$pkg}{'branch'});
  $$versions{$pkgs->{'pkgs'}{$pkg}{'ppa'}}    = get_pkg_version($pkgs->{'pkgs'}{$pkg}{'ppa'});
  do_revert($pkgs->{'pkgs'}{$pkg}{'ppa'},
            $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}}{'version'}, undef)
    if $revert && defined $$rpkgs{$pkg};
  # check the dist branches too
  for my $d ($conf->{'dist'}, @{$conf->{'alt_dists'}}) {
    $pkgs->{'pkgs'}{$pkg}{"ppa.$d"} = $pkgs->{'pkgs'}{$pkg}{"ppa"} . ".$d";
    next unless -d $pkgs->{'pkgs'}{$pkg}{"ppa.$d"};
    check_committed($pkg, $pkgs->{'pkgs'}{$pkg}{"ppa.$d"}) if defined $$rpkgs{$pkg} && !$revert;
    $$versions{$pkgs->{'pkgs'}{$pkg}{"ppa.$d"}} = get_pkg_version($pkgs->{'pkgs'}{$pkg}{"ppa.$d"});
    do_revert($pkgs->{'pkgs'}{$pkg}{"ppa.$d"},
              $$versions{$pkgs->{'pkgs'}{$pkg}{"ppa"}}{'version'},
              $$versions{$pkgs->{'pkgs'}{$pkg}{"ppa.$d"}}{'version'})
      if $revert && defined $$rpkgs{$pkg};
  }
}

exit 0 if $revert;

for my $pkg (@{$pkgs->{'order'}}) {
  next unless defined $pkgs->{'pkgs'}{$pkg};
  do_pkg($pkg, $force);
}

if ($#$dput >= 0) {
  print "\n\n# Please double check the results then issue:\ncd $build_dir; " .
  "dput $dput_dist \\\n	" . (join " \\\n	", @$dput) . "\n";
  if ($conf->{'do_dput'}) {
    chdir($build_dir) || die "Can't chdir to $build_dir: $!\n";
    if (0) {
      print "# workaround for LP #371640\n";
      my @orig = grep {/~$conf->{'iteration_tag'}1_/} @$dput;
      @$dput = grep {!/~$conf->{'iteration_tag'}1_/} @$dput;
      system("dput", $dput_dist, @orig) if $#orig >= 0;
      if ($#orig >= 0 && $#$dput >= 0) {
        print "# Sleeping for 10 minutes...\n";
        sleep 600;
      }
    }
    if ($#$dput >= 0) {
      system("dput", $dput_dist, @$dput) == 0 && do {
        print "*LOG* dput successfully pushed all " . ($#$dput + 1) . " source packages (to $dput_dist)\n";
      } || do {
        print "*LOG* dput '$dput_dist' returned with error $?: $!\n";
        # workaround LP regression. Only the 1st has been uploaded
        for my $file (@$dput[1..$#$dput]) {
          system("dput", $dput_dist, $file); # ignore errors, they are expected here :(
        }
        for my $file (@$dput) {
          $file =~ s/(.*)\.changes$/$1\.$dput_dist.upload/;
          printf "*LOG* dput: %s not uploaded\n", $1 unless -f $file;
        }
      };
    }
  }
}

# End

sub check_committed {
  my $pkg    = shift;
  my $branch = shift;

  print "* check_committed($branch)\n";
  chdir($branch) || die "Can't chdir to $branch: $!\n";
  system("bzr", "diff") == 0 || die "*LOG* sync $branch: this branch has uncommitted changes. Abort";
  # Check for new revisions
  my @args = ("pull");
  push @args, $pkgs->{'pkgs'}{$pkg}{'lp'} if defined $pkgs->{'pkgs'}{$pkg}{'lp'};
  system("bzr", @args) unless $branch =~ m/$conf->{'ppa_branchtag'}($|\.\S+$)/;
  my $id = `bzr revno`;
  chomp($id);
  $$cids{$branch} = $id;
  chdir("..");
}

sub get_pkg_version {
  my $pkg = shift;

  open(FILE, "$pkg/debian/changelog") || do {
    warn "Can't open $pkg/debian/changelog: $!\n";
    return {};
  };
  my $line = <FILE>;
  $line =~ m,^\S+ \((\S+)\) (\S+); urgency=\S+,;
  close FILE;
  return { 'version' => $1, 'dist' => $2 };
}

sub do_revert {
  my $branch  = shift;
  my $vbranch = shift;
  my $vppa    = shift;

  print "do_revert($branch)\n";
  $vbranch =~ s,([\~\+]),\\$1,g;
  if (!defined $vppa || $vppa =~ m/^$vbranch\~/) {
    chdir($branch) || die "Can't chdir to $branch: $!\n";
    print "*INFO* do revert $branch\n";
    system("bzr", "uncommit", "--force") == 0 || die "Branch $branch cannot be uncommitted. Stop";
    system("bzr", "revert") == 0 || die "Branch $branch cannot be reverted. Stop";
    chdir("..");
  }
  else {
    print "*INFO* not reverting $branch: $vppa !~ m/^$vbranch\~/\n";
  }
}

sub do_test_build {
  my $pkg    = shift;
  my $native = shift;

  # bzr bdm / bdn
  if ($native) {
    my $cmd = [ "bzr", @{$$bzr_cmds{'bdn'}}, "-e" ];
    print "\$ " . (join ' ', @$cmd) . " (bdn)\n";
    system(@$cmd);
  }
  else {
    my $cmd = [ "bzr", @{$$bzr_cmds{'bdm'}}, "-e" ];
    print "\$ " . (join ' ', @$cmd) . " (bdm)\n";
    system(@$cmd);
  }
  # TODO: enter the build dir and run pre-build/patch
  0
}

sub do_merges {
  my $pkg      = shift;
  my $force    = shift;
  my $target   = shift;
  my $suffix   = shift;
  my $isclosed = shift;
  my $new_orig = shift;
  my $native   = shift;

  if ($isclosed) {
    $target =~ s/(\d+)$//;
    $target .= $new_orig ? 1 : $1 + 1;
  }

  my $ret = do_merge($pkg, $force, $target, $isclosed, $new_orig, $native, 0, $conf->{'dist'});
  if ($ret > 0) {
    print "Error during merge? Investigate. Skip this package.\n";
    return;
  }
  return if $ret == -1;
  for my $d ($conf->{'dist'}, @{$conf->{'alt_dists'}}) {
    do_merge($pkg, $force, $target . $suffix, 0,
             $d eq $conf->{'dist'} ? $new_orig : 0, $native, 1, $d);
  }
}

sub do_merge {
  my $pkg      = shift;
  my $force    = shift;
  my $target   = shift;
  my $isclosed = shift;
  my $new_orig = shift;
  my $native   = shift;
  my $merge    = shift;
  my $ldist    = shift;

  my $ppa = $pkgs->{'pkgs'}{$pkg}{'ppa'};
  my $ppaname = 'ppa';
  my $merge_from = 'branch';
  if ($merge) {
    $ppa .= ".$ldist";
    $ppaname .= ".$ldist";
    $target .= "~$ldist" unless $ldist eq $conf->{'dist'};
    $merge_from = 'ppa';
  }
  print "\$ cd $ppa\n";
  chdir($ppa) || die "Can't chdir to $ppa: $!\n";
  if (0 && $ldist eq $conf->{'dist'}) {
    # Perform some minimal tests on the branch
    my $ret = do_test_build();
    if ($ret) {
      print "Caught something. Abort\n";
      return $ret;
    }
  }
  print "\$ bzr merge ../$pkgs->{'pkgs'}{$pkg}{$merge_from}\n";
  system("bzr", "merge", "../$pkgs->{'pkgs'}{$pkg}{$merge_from}");
  my $lines = `bzr st | sed -e '/^unknown:\$/,/^[ ]/d'`; # ignore possible leftovers
  print "\$ bzr st\n$lines";
  if ($lines eq '') {
    print "    => nothing to merge" . ($force ? " but force requested" : "") . "\n";
    print "    => ppa = $$versions{$ppa}{'version'}\n";
    unless ($merge || $force) {
      print "*LOG* sync $pkg: nothing to merge from $pkgs->{'pkgs'}{$pkg}{$merge_from}\n";
      chdir("..");
      return $new_orig ? 0 : -1;
    }
    $new_orig = 1 if $ldist eq $conf->{'dist'};
  }
  my $conflicts = {};
  my @ls = split /\n/, $lines;
  for my $line (@ls) {
    $$conflicts{$1} = 1 if $line =~ m/Text conflict in (\S+)/;
  }
  $$versions{$ppa}{'version'} = $target;

  # Update debian/changelog
  if (defined $$conflicts{"debian/changelog"}) {
    print "\$ bzr revert debian/changelog\n";
    system("bzr", "revert", "debian/changelog");
    delete $$conflicts{"debian/changelog"};
  }
  print "\$ cp ../$pkgs->{'pkgs'}{$pkg}{$merge_from}/debian/changelog debian/changelog\n";
  `cp ../$pkgs->{'pkgs'}{$pkg}{$merge_from}/debian/changelog debian/changelog`;
  my $dist = $merge ? $ldist : "UNRELEASED";
  if ($isclosed) {
    # Note: we have to use -b as we are always lower
    print "\$ DEBEMAIL=\"$conf->{'who'}\" dch -b -v $target -D $dist \"New upstream snapshot\"\n";
    `DEBEMAIL="$conf->{'who'}" dch -b -v $target -D $dist "New upstream snapshot"`;
  }
  else {
    print "TARGET=$target\n";
    print "\$ sed -i -e '1s/(.*) [^;]*/($target) $dist/' debian/changelog\n";
    `sed -i -e '1s/(.*) [^;]*/($target) $dist/' debian/changelog`;
    my $maint = `grep '^ -- ' debian/changelog | head -1 | sed -e 's/^ -- \(.*\)  .*/\1/'`;
    if ($maint eq $conf->{'who'}) {
      print "\$ sed -i -e '1,/^ -- /s/^ -- .*/ -- $conf->{'who'}  $date/' debian/changelog\n";
      `sed -i -e '1,/^ -- /s/^ -- .*/ -- $conf->{'who'}  $date/' debian/changelog`;
    }
    elsif (! $merge) {
      # let dch take care of giving credits to all maintainers
      print "\$ DEBEMAIL=\"$conf->{'who'}\" dch -D $dist \"\"\n";
      `DEBEMAIL="$conf->{'who'}" dch -D $dist ""`;
      my $nmaint = `grep '^ -- ' debian/changelog | head -1 | sed -e 's/^ -- \(.*\) <.*>  .*/\1/'`;
      `perl -pe 'BEGIN { \$s = "$nmaint"; \$/ = undef; } s/^  \\[ \$s \\]\\n\\n( -- \$s <)/\$1/m' debian/changelog`;
    }
  }
  if (defined $$conflicts{'debian/changelog'}) {
    print "\$ bzr resolve debian/changelog\n";
    system("bzr", "resolve", "debian/changelog");
  }

  # Update debian/control
  print "\$ bzr revert debian/control\n";
  system("bzr", "revert", "debian/control");
  print "\$ cp ../$pkgs->{'pkgs'}{$pkg}{$merge_from}/debian/control debian/control\n";
  `cp ../$pkgs->{'pkgs'}{$pkg}{$merge_from}/debian/control debian/control`;
  my $dodeps = 1;
  for my $p (@{$pkgs->{'nodeps'}}) {
    $dodeps = 0, last if $p eq $pkg;
  }
  if ($dodeps) {
    for my $dep (keys %{$pkgs->{'deps'}}) {
      next unless defined $pkgs->{'deps'}{$dep} && defined $pkgs->{'pkgs'}{$pkgs->{'deps'}{$dep}};
      my $vers = $$versions{$pkgs->{'pkgs'}{$pkgs->{'deps'}{$dep}}{$ppaname}}{'version'};
      if (defined $vers) {
        print "\$ sed -i -e '1,/^Package: /s/\\(${dep}\\)[^,]*/\\1 (>= $vers)/' debian/control\n";
        `sed -i -e '1,/^Package: /s/\\($dep\\)[^,]*/\\1 (>= $vers)/' debian/control`;
      }
      else {
        print "# not updating dep for $dep, version is missing\n";
      } 
    }
  }
  my $bdeps = [];
  if ($merge && defined $pkgs->{'build-deps'}) { # Global 'build-deps' structure
    for my $key (keys %{$pkgs->{'build-deps'}}) {
      ($key eq 'all_pkgs' || $key eq $pkg) && do {
        push @$bdeps, @{$pkgs->{'build-deps'}{$key}}
      };    
    }
  }
  if ($merge && defined $pkgs->{'pkgs'}{$pkg}{'build-deps'}) {
    for my $key (keys %{$pkgs->{'pkgs'}{$pkg}{'build-deps'}}) {
      ($key eq 'all_dists' || $key eq $ldist) && do {
        push @$bdeps, @{$pkgs->{'pkgs'}{$pkg}{'build-deps'}{$key}}
      };
    }
  }
  if ($#$bdeps >= 0) {
   # Tweak Build-Depends according to the 'build-deps' preferences in that pocket
   my $v = join ', ', @$bdeps;
   print "**INFO** Build-Depends += $v\n";
   `sed -i -e '1,/^Package: /s/\\(Build-Depends:[ 	]*\\)/\\1$v, /' debian/control`;
  }
  if (defined $$conflicts{'debian/control'}) {
    print "\$ bzr resolve debian/control\n";
    system("bzr", "resolve", "debian/control");
  }
  # `ls -l debian/control.~*~ debian/changelog.~*~`;
  print "\$ bzr diff\n";
  system("bzr", "diff");
  if ($merge) {
    print "\$ bzr commit -m \"* RELEASE $target to $ldist/ppa\\n" .
          "  - merge with $pkgs->{'pkgs'}{$pkg}{$merge_from} #" . $$cids{$pkgs->{'pkgs'}{$pkg}{$merge_from}} . "\"\n";
    system("bzr", "commit", "-m", "* RELEASE $target to $ldist/ppa\n" .
           "  - merge with $pkgs->{'pkgs'}{$pkg}{$merge_from} \#" . $$cids{$pkgs->{'pkgs'}{$pkg}{$merge_from}});
  }
  else {
    print "\$ bzr commit -m \"* Merge with $pkgs->{'pkgs'}{$pkg}{$merge_from} #" . $$cids{$pkgs->{'pkgs'}{$pkg}{$merge_from}} . "\"\n";
    system("bzr", "commit", "-m", "* Merge with $pkgs->{'pkgs'}{$pkg}{$merge_from} \#" . $$cids{$pkgs->{'pkgs'}{$pkg}{$merge_from}});
    print "*LOG* sync $pkg: merged rev #$$cids{$pkgs->{'pkgs'}{$pkg}{$merge_from}} from $pkgs->{'pkgs'}{$pkg}{$merge_from}\n";
  }
  chomp(my $revno = `bzr revno`);
  $$cids{$pkgs->{'pkgs'}{$pkg}{$ppaname}} = $revno;
  print "    => committed $target in $pkgs->{'pkgs'}{$pkg}{$ppaname}\n";
  unless ($merge) {
    chdir("..");
    return 0;
  }
  if ($native) {
    my $cmd = [ "bzr", @{$$bzr_cmds{'ppan'}} ];
    print "\$ " . (join ' ', @$cmd) . " (ppan)\n";
    system(@$cmd);
  }
  else {
    my $target = ($new_orig || ($force && $ldist eq $conf->{'dist'})) ? "ppa" : "ppa2";
    my $cmd = [ "bzr", @{$$bzr_cmds{$target}} ];
    print "\$ " . (join ' ', @$cmd) . " #($target)\n";
    system(@$cmd);
  }
  chdir("..");
  my $version = $target;
  $version =~ s/^[^:]*://;
  push @$dput, "${pkg}_${version}_source.changes";
  0;
}

sub do_pkg {
  my $pkg   = shift;
  my $force = shift;

  my $target;
  print "* Doing $pkg from $pkgs->{'pkgs'}{$pkg}{'branch'} to $pkgs->{'pkgs'}{$pkg}{'ppa'}\n";
  printf "  - compare %s (%s) and %s\n", $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}}{'version'},
   $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}}{'dist'}, $$versions{$pkgs->{'pkgs'}{$pkg}{'ppa'}}{'version'};
  my $isclosed = $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}}{'dist'} eq 'UNRELEASED' ? 0 : 1;
  my $isnative = $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}}{'version'} =~ m/-/ ? 0 : 1;
  my $isahead = `dpkg --compare-versions $$versions{$pkgs->{'pkgs'}{$pkg}{'branch'}}{'version'} gt $$versions{$pkgs->{'pkgs'}{$pkg}{'ppa'}}{'version'}; [ \$? -eq 0 ] && echo 1 || echo 0`;
  chomp($isahead);
  print "     => isclosed=$isclosed isnative=$isnative isahead=$isahead\n";
  printf "  - lastest daily version: %s\n", $$versions{$pkgs->{'pkgs'}{$pkg}{'ppa.' . $conf->{'dist'}}}{'version'};
  my $from = $isahead ? 'branch' : 'ppa';
  my $to   = 'ppa.' . $conf->{'dist'};
  print "     => from=$from to=$to\n";
  # base_branch refers to the branch updated by the update script
  my ($base_branch, $iter_branch, $base_ppa, $iter_ppa);
  if ($isnative) {
    ($base_branch, $iter_branch) = $$versions{$pkgs->{'pkgs'}{$pkg}{$from}}{'version'} =~ m/^(\S+?)(~.*|$)/;
    $iter_branch =~ s/\~$conf->{'iteration_tag'}(\d+)$//;
    ($base_ppa, $iter_ppa) = $$versions{$pkgs->{'pkgs'}{$pkg}{$to}}{'version'} =~ m/^(\S+?)(~.*|$)/;
  }
  else {
    ($base_branch, $iter_branch) = $$versions{$pkgs->{'pkgs'}{$pkg}{$from}}{'version'} =~ m/^(\S+)-(\S+)/;
    $iter_branch =~ s/\~$conf->{'iteration_tag'}(\d+)$//;
    ($base_ppa, $iter_ppa) = $$versions{$pkgs->{'pkgs'}{$pkg}{$to}}{'version'} =~ m/^(\S+)-(\S+)/;
  }
  print "     => base_branch=$base_branch iter_branch=$iter_branch\n";
  print "     => base_ppa=$base_ppa iter_ppa=$iter_ppa\n";
  if ($isnative) {
    $iter_ppa =~ s/\~$conf->{'iteration_tag'}(\d+)/$1/;
    my $id = $1;
    print "    => need to check the branches :P\n";
    $id = 0 if $base_branch ne $base_ppa || !defined $id;
    do_merges($pkg, $force, $base_branch,
              "~$conf->{'iteration_tag'}" . ($id + 1), $isclosed, 0, 1);
  }
  elsif ($base_branch ne $base_ppa) {
    print "    => upstream version differs ($base_branch vs $base_ppa): continue\n";
    do_merges($pkg, $force, $$versions{$pkgs->{'pkgs'}{$pkg}{$from}}{'version'},
              "~$conf->{'iteration_tag'}" . 1, $isclosed, 1, 0);
  }
  else {
    $iter_ppa =~ s/(\S+)\~$conf->{'iteration_tag'}(\d+)/$1/;
    my ($iter_ppa2, $id) = ($1, $2);
    if (defined $iter_ppa2 && $iter_branch ne $iter_ppa2) {
      print "     => same upstream version but ubuntu version differs " . 
            "($iter_branch vs $iter_ppa2): continue\n";
      do_merges($pkg, $force, $$versions{$pkgs->{'pkgs'}{$pkg}{$from}}{'version'},
                "~$conf->{'iteration_tag'}" . 1, $isclosed, 0, 0);
    }
    else {
      $id = 0 unless defined $id;
      print "    => need to check the branches :P\n";
      do_merges($pkg, $force, $base_ppa . "-" . $iter_ppa,
                "~$conf->{'iteration_tag'}" . ($id + 1), $isclosed, 0, 0);
    }
  }
}

sub read_conf {
  my $conf   = shift;
  my $file   = shift;
  my $strict = shift;

  local $/ = undef;
  open(CONF, "$file") || die "Can't open $file: $!\n";
  my $data = <CONF>;
  close CONF;
  my $newconf = eval $data;
  die "Syntax error in $file? You can check your syntax with 'perl -c $file'\n" unless defined $newconf;

  for my $key (keys %$newconf) {
    if ($strict && !exists $$$conf{$key}) {
      warn "Unrecognized conf '$key' found in $file: ignored\n";
      next; 
    }
    if (exists $$$conf{$key} && $$$conf{$key} =~ /_dir$/) {
        $$$conf{$key} = File::Spec->rel2abs($$newconf{$key});
        $$$conf{$key} = File::Spec->rel2abs($$newconf{$key});
    } else {
        $$$conf{$key} = $$newconf{$key};
    }
  }
}
